@@ -22,6 +22,8 @@ import matplotlib.pyplot as plt
 from triqler import qvality
 
 from . import qvalues
+from . import utils
+from .picked_protein import picked_protein
 
 LOGGER = logging.getLogger(__name__)
 
@@ -59,11 +61,12 @@ class Confidence:
         self._data[self._score_column] = scores * (desc * 2 - 1)
 
         # This attribute holds the results as DataFrames:
-        self._confidence_estimates = {}
+        self.confidence_estimates = {}
+        self.decoy_confidence_estimates = {}
 
     def __getattr__(self, attr):
         try:
-            return self._confidence_estimates[attr]
+            return self.confidence_estimates[attr]
         except KeyError:
             raise AttributeError
 
@@ -72,9 +75,9 @@ class Confidence:
         """
         The available levels for confidence estimates.
         """
-        return list(self._confidence_estimates.keys())
+        return list(self.confidence_estimates.keys())
 
-    def to_txt(self, dest_dir=None, file_root=None, sep="\t"):
+    def to_txt(self, dest_dir=None, file_root=None, sep="\t", decoys=False):
         """
         Save confidence estimates to delimited text files.
 
@@ -89,6 +92,8 @@ class Confidence:
             `mokapot.peptides.txt`.
         sep : str
             The delimiter to use.
+        decoys : bool
+            Save decoys confidence estimates as well?
 
         Returns
         -------
@@ -102,11 +107,23 @@ class Confidence:
             file_base = os.path.join(dest_dir, file_base)
 
         out_files = []
-        for level, qvals in self._confidence_estimates.items():
+        for level, qvals in self.confidence_estimates.items():
+            if qvals is None:
+                continue
+
             out_file = file_base + f".{level}.txt"
             qvals.to_csv(out_file, sep=sep, index=False)
             out_files.append(out_file)
 
+        if decoys:
+            for level, qvals in self.decoy_confidence_estimates.items():
+                if qvals is None:
+                    continue
+
+                out_file = file_base + f"decoys.{level}.txt"
+                qvals.to_csv(out_file, sep=sep, index=False)
+                out_files.append(out_file)
+
         return out_files
 
     def _perform_tdc(self, psm_columns):
@@ -118,7 +135,9 @@ class Confidence:
         psm_columns : str or list of str
             The columns that define a PSM.
         """
-        psm_idx = _groupby_max(self._data, psm_columns, self._score_column)
+        psm_idx = utils.groupby_max(
+            self._data, psm_columns, self._score_column
+        )
         self._data = self._data.loc[psm_idx, :]
 
     def plot_qvalues(self, level="psms", threshold=0.1, ax=None, **kwargs):
@@ -146,12 +165,11 @@ class Confidence:
             An :py:class:`matplotlib.axes.Axes` with the cumulative
             number of accepted target PSMs or peptides.
         """
-        ax = plot_qvalues(
-            self._confidence_estimates[level]["mokapot q-value"],
-            threshold=threshold,
-            ax=ax,
-            **kwargs,
-        )
+        qvals = self.confidence_estimates[level]["mokapot q-value"]
+        if qvals is None:
+            raise ValueError(f"{level}-level estimates are unavailable.")
+
+        ax = plot_qvalues(qvals, threshold=threshold, ax=ax, **kwargs)
         ax.set_xlabel("q-value")
         ax.set_ylabel(f"Accepted {self._level_labs[level]}")
 
@@ -184,6 +202,12 @@ class LinearConfidence(Confidence):
         Confidence estimates for PSMs in the dataset.
     peptides : pandas.DataFrame
         Confidence estimates for peptides in the dataset.
+    proteins : pandas.DataFrame or None
+        Confidence estimates for proteins in the dataset.
+    confidence_estimates : Dict[str, pandas.DataFrame]
+        A dictionary of confidence estimates at each level.
+    decoy_confidence_estimates : Dict[str, pandas.DataFrame]
+        A dictionary of confidence estimates for the decoys at each level.
     """
 
     def __init__(self, psms, scores, desc=True, eval_fdr=0.01):
@@ -230,8 +254,11 @@ class LinearConfidence(Confidence):
 
     def _num_accepted(self, level):
         """Calculate the number of accepted discoveries"""
-        disc = self._confidence_estimates[level]
-        return (disc["mokapot q-value"] <= self._eval_fdr).sum()
+        disc = self.confidence_estimates[level]
+        if disc is not None:
+            return (disc["mokapot q-value"] <= self._eval_fdr).sum()
+        else:
+            return None
 
     def _assign_confidence(self, desc=True):
         """
@@ -243,7 +270,7 @@ class LinearConfidence(Confidence):
             Are higher scores better?
         """
         levels = ["PSMs", "peptides"]
-        peptide_idx = _groupby_max(
+        peptide_idx = utils.groupby_max(
             self._data, self._peptide_column, self._score_column
         )
 
@@ -253,7 +280,13 @@ class LinearConfidence(Confidence):
         level_data = [self._data, peptides]
 
         if self._has_proteins:
-            proteins = self._picked_protein(peptides)
+            proteins = picked_protein(
+                peptides,
+                self._target_column,
+                self._peptide_column[0],
+                self._score_column,
+                self._proteins,
+            )
             levels += ["proteins"]
             level_data += [proteins]
             LOGGER.info("\t- Found %i unique protein groups.", len(proteins))
@@ -264,12 +297,11 @@ class LinearConfidence(Confidence):
 
             # Estimate q-values and assign to dataframe
             LOGGER.info("Assiging q-values to %s...", level)
-            data["mokapot q-value"] = qvalues.tdc(scores, targets, desc)
+            data["mokapot q-value"] = qvalues.tdc(scores, targets, desc=True)
 
             # Make output tables pretty
             data = (
-                data.loc[targets, :]
-                .sort_values(self._score_column, ascending=(not desc))
+                data.sort_values(self._score_column, ascending=False)
                 .reset_index(drop=True)
                 .drop(self._target_column, axis=1)
                 .rename(columns={self._score_column: "mokapot score"})
@@ -289,81 +321,17 @@ class LinearConfidence(Confidence):
             # Calculate PEPs
             LOGGER.info("Assiging PEPs to %s...", level)
             _, pep = qvality.getQvaluesFromScores(
-                scores[targets], scores[~targets]
+                scores[targets], scores[~targets], includeDecoys=True
             )
+
+            level = level.lower()
             data["mokapot PEP"] = pep
-            self._confidence_estimates[level.lower()] = data
+            self.confidence_estimates[level] = data.loc[targets, :]
+            self.decoy_confidence_estimates[level] = data.loc[~targets, :]
 
-    def _picked_protein(self, peptides):
-        """
-        Calculate protein-level confidence estimates.
-
-        Returns
-        -------
-        pandas.DataFrame
-            The protein confidence estimates.
-        """
-        keep = [
-            self._target_column,
-            self._peptide_column[0],
-            self._score_column,
-        ]
-
-        prots = peptides.loc[:, keep].rename(
-            columns={self._peptide_column[0]: "best peptide"}
-        )
-
-        prots["stripped sequence"] = (
-            prots["best peptide"]
-            .str.replace(r"[\[\(].*?[\]\)]", "")
-            .str.replace(r"^.*?\.", "")
-            .str.replace(r"\..*?$", "")
-        )
-
-        prots["mokapot protein group"] = prots["stripped sequence"].map(
-            self._proteins.peptide_map.get
-        )
-
-        unmatched = pd.isna(prots["mokapot protein group"])
-        unmatched_prots = prots.loc[unmatched, :]
-        shared = unmatched_prots["stripped sequence"].isin(
-            self._proteins.shared_peptides
-        )
-
-        shared_unmatched = (~shared).sum()
-        num_shared = len(shared) - shared_unmatched
-        LOGGER.debug(
-            "%i out of %i peptides were discarded as shared peptides.",
-            num_shared,
-            len(prots),
-        )
-
-        if shared_unmatched:
-            LOGGER.warning(
-                "%i out of %i peptides could not be mapped."
-                "Check your digest settings.",
-                shared_unmatched,
-                len(prots),
-            )
-
-        prots = prots.loc[~unmatched, :]
-
-        prots["decoy"] = (
-            prots["mokapot protein group"]
-            .str.split(",", expand=True)[0]
-            .map(lambda x: self._proteins.protein_map.get(x, x))
-        )
-
-        prot_idx = _groupby_max(prots, ["decoy"], self._score_column)
-        final_cols = [
-            "mokapot protein group",
-            "best peptide",
-            "stripped sequence",
-            self._score_column,
-            self._target_column,
-        ]
-
-        return prots.loc[prot_idx, final_cols]
+        if "proteins" not in self.confidence_estimates.keys():
+            self.confidence_estimates["proteins"] = None
+            self.decoy_confidence_estimates["proteins"] = None
 
 
 class CrossLinkedConfidence(Confidence):
@@ -411,7 +379,7 @@ class CrossLinkedConfidence(Confidence):
         desc : bool
             Are higher scores better?
         """
-        peptide_idx = _groupby_max(
+        peptide_idx = utils.groupby_max(
             self._data, self._peptide_columns, self._score_column
         )
 
@@ -495,18 +463,6 @@ def plot_qvalues(qvalues, threshold=0.1, ax=None, **kwargs):
     return ax
 
 
-def _groupby_max(df, by_cols, max_col):
-    """Quickly get the indices for the maximum value of col"""
-    idx = (
-        df.sample(frac=1)
-        .sort_values(list(by_cols) + [max_col], axis=0)
-        .drop_duplicates(list(by_cols), keep="last")
-        .index
-    )
-
-    return idx
-
-
 def _new_column(name, df):
     """Add a new column, ensuring a unique name"""
     new_name = name
